package org.apache.xindice.core.filer;

/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * $Id: MemFiler.java,v 1.1 2001/12/06 21:00:12 bradford Exp $
 */

import org.apache.xindice.core.*;
import org.apache.xindice.core.Collection;
import org.apache.xindice.core.data.*;
import org.apache.xindice.util.*;

import java.util.*;

/**
 * MemFiler is an In-Memory Filer implementation for Xindice.  MemFiler can be
 * used for temporary collections and caching.  It's basically a layering on
 * top of HashMap.
 */

public final class MemFiler extends SimpleConfigurable implements Filer {
   private Map hashTable = null;
   private boolean opened = false;
   private boolean readOnly = false;
   private Collection collection;

   public MemFiler() {
      hashTable = Collections.synchronizedMap(new HashMap());
   }

   public MemFiler(Map hashTable, boolean readOnly) {
      this.hashTable = hashTable;
      this.readOnly = readOnly;
   }
   
   public MemFiler(Map hashTable) {
      this(hashTable, false);
   }
   
   public void setCollection(Collection collection) {
      this.collection = collection;
   }

   public String getName() {
      return "MemFiler";
   }

   private void checkOpened() throws DBException {
      if ( !opened )
         throw new FilerException(FaultCodes.COL_COLLECTION_CLOSED, "Filer is closed");
   }

   private void checkReadOnly() throws DBException {
      if ( readOnly )
         throw new FilerException(FaultCodes.COL_COLLECTION_READ_ONLY, "Filer is read-only");
   }
   
   public boolean create() {
      hashTable.clear();
      return true;
   }

   public boolean open() {
      opened = true;
      return opened;
   }

   public boolean isOpened() {
      return opened;
   }

   public boolean exists() {
      return true;
   }

   public boolean drop() {
      hashTable.clear();
      opened = false;
      return !opened;
   }

   public boolean close() {
      opened = false;
      return !opened;
   }
   
   public void flush() {
   }
   
   public Record readRecord(Key key) throws DBException {
      checkOpened();
      return(Record)hashTable.get(key);
   }

   public boolean writeRecord(Key key, Value value) throws DBException {
      checkOpened();
      checkReadOnly();
      hashTable.put(key, new Record(key, value));
      return true;
   }

   public boolean deleteRecord(Key key) throws DBException {
      checkOpened();
      checkReadOnly();
      return hashTable.remove(key) != null;
   }

   public long getRecordCount() throws DBException {
      checkOpened();
      return hashTable.size();
   }

   public RecordSet getRecordSet() throws DBException {
      checkOpened();
      return new MemRecordSet();
   }

   
   /**
    * MemRecordSet
    */
   
   private class MemRecordSet implements RecordSet {
      private Iterator enum = hashTable.values().iterator();

      public boolean hasMoreRecords() throws DBException {
         return enum.hasNext();
      }

      public Record getNextRecord() throws DBException {
         checkOpened();
         return(Record)enum.next();
      }

      public Value getNextValue() throws DBException {
         checkOpened();
         return((Record)enum.next()).getValue();
      }
      
      public Key getNextKey() {
         return((Record)enum.next()).getKey();
      }
   }
}


