package org.apache.xindice.server.standard;

/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * $Id: StdFileManager.java,v 1.1.1.1 2001/12/06 19:33:56 bradford Exp $
 */

import org.apache.xindice.server.*;
import org.apache.xindice.util.*;
import java.io.*;
import java.util.*;

/**
 * StdFileManager
 */

public final class StdFileManager implements Configurable, FileManager, KernelAccess {
   private static final String MINBYTES = "minbytes";
   private static final String MAXBYTES = "maxbytes";
   private static final String CHUNKSIZE = "chunksize";
   private static final String TMPDIR = "tmpdir";

   private static final String TMPEXT = ".tmp";

   private Kernel kernel;
   private Configuration config;
   private FileCache filecache = new FileCache();
   private int minBytes = 0;       // 0k
   private int maxBytes =  131072; // 128k
   private int chunkSize = 32768;  // 32k
   private File tempDir;
   private long counter = this.hashCode();

   public void setKernel(Kernel kernel) {
      this.kernel = kernel;
   }

   public void setConfig(Configuration config) {
      this.config = config;
      minBytes = config.getIntAttribute(MINBYTES);
      maxBytes = config.getIntAttribute(MAXBYTES);
      chunkSize = config.getIntAttribute(CHUNKSIZE);
      tempDir = new File(config.getAttribute(TMPDIR));
   }

   public Configuration getConfig() {
      return config;
   }

   private synchronized String getTempFilename() {
      return StringUtilities.rightJustify(Long.toHexString(counter++), 8).replace(' ', '0')+TMPEXT;
   }

   public FileResource getFileResource(String filename) {
      return new StdFileResource(filename);
   }

   public FileResource getTemporaryFileResource() {
      return new StdFileResource(tempDir, getTempFilename(), true);
   }

   private class StdFileResource extends File implements FileResource {
      private boolean temporary = false;

      public StdFileResource(String filename, boolean temporary) {
         super(filename);
         this.temporary = temporary;
      }

      public StdFileResource(String filename) {
         super(filename);
      }

      private StdFileResource(File directory, String name, boolean temporary) {
         super(directory, name);
         this.temporary = temporary;
      }

      private StdFileResource(File directory, String name) {
         super(directory, name);
      }

      public InputStream getInputStream() {
         try {
            if ( (length() >= minBytes) && (length() <= maxBytes) )
               return new ByteArrayInputStream(filecache.getFile(getCanonicalPath()));
            else
               return new FileInputStream(this);
         }
         catch ( Exception e ) {
            System.err.println("\u0007ERROR: "+e);
         }
         return null;
      }

      public OutputStream getOutputStream() {
         try {
            return new FileOutputStream(this);
         }
         catch ( Exception e ) {
            System.err.println("\u0007ERROR: "+e);
         }
         return null;
      }

      public OutputStream getAppendStream() {
         try {
            return new FileOutputStream(this.getAbsolutePath(), true);
         }
         catch ( Exception e ) {
            System.err.println("\u0007ERROR: "+e);
         }
         return null;
      }

      public void copyStreamTo(OutputStream os) throws IOException {
         int size = 0;
         byte[] buffer;
         if ( (length() >= minBytes) && (length() <= maxBytes) ) {
            buffer = filecache.getFile(getCanonicalPath());
            int pos = 0;
            int length = buffer.length;
            while ( pos < length ) {
               size = length-pos > chunkSize ? chunkSize
                                             : length-pos;
               os.write(buffer, pos, size);
               pos += size;
               Thread.yield();
            }
         }
         else {
            InputStream is = null;
            try {
               is = new FileInputStream(this);
               buffer = new byte[chunkSize];
               while ( is.available() > 0 )
                  if ( (size = is.read(buffer)) > 0 ) {
                     os.write(buffer, 0, size);
                     Thread.yield();
                  }
            }
            finally {
               is.close();
            }
         }
      }

      public boolean move(String filename) {
         return renameTo(new File(filename));
      }

      public boolean rename(String filename) {
         return renameTo(new File(filename));
      }

      public FileResource[] getChildren() {
         String[] files = super.list();
         FileResource[] children = new StdFileResource[files.length];
         for ( int i = 0; i < files.length; i++ )
            children[i] = new StdFileResource(this, files[i]);
         return children;
      }

      protected void finalize() throws Throwable {
         if ( temporary )
            delete();
      }
   }
}
