package org.apache.xindice.util;

/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * $Id: Configuration.java,v 1.2 2002/02/24 01:20:27 bradford Exp $
 */

import org.w3c.dom.*;

import java.util.*;
import java.io.*;

/**
 * Configuration is a utility class that allows Configurable objects to
 * easily manage configuration information in a simple and consistent
 * fashion.
 *
 * @see org.apache.xindice.util.Configurable
 */

public final class Configuration {
   private static final Configuration[] EmptySet = new Configuration[0];
   
   public static final Map Modified = Collections.synchronizedMap(new HashMap()); // Temporary HACK

   private Element config = null;
   private boolean readOnly = true;

   public Configuration(Element config, boolean readOnly) {
      this.config = config;
      this.readOnly = readOnly;
   }
   
   public Configuration(Document config, boolean readOnly) {
      this(config.getDocumentElement(), readOnly);
   }

   public Configuration(Element config) {
      this(config, true);
   }
   
   public Configuration(Document config) {
      this(config.getDocumentElement());
   }

   /**
    * getElement returns the Element being managed by this
    * Configuration.
    *
    * @return The Configuration Element
    * @throws ReadOnlyException If the Configuration is Read-only
    */
   public Element getElement() throws ReadOnlyException {
      if ( readOnly )
         throw new ReadOnlyException();
      return config;
   }
   
   /**
    * getName returns the name of the Configuration node.
    *
    * @return The Node name
    */
   public String getName() {
      return config.getNodeName();
   }

   /**
    * hasAttributes returns whether or not the Configuration node has any
    * attributes associated with it.
    *
    * @return Whether or not attributes exist
    */
   public boolean hasAttributes() {
      return config.getAttributes().getLength() > 0;
   }

   public void setDirty() {
      Modified.put(config.getOwnerDocument(), Boolean.TRUE);
   }

   /**
    * getAttribute returns an attribute from the Configuration node.
    *
    * @param name The attribute name
    * @param defValue The default value
    * @return The attribute value
    */
   public String getAttribute(String name, String defValue) {
      String value = config.getAttribute(name);
      if ( value == null || value.length() == 0 )
         value = defValue;
      return value;
   }

   /**
    * getAttribute returns an attribute from the Configuration node.
    *
    * @param name The attribute name
    * @return The attribute value
    */
   public String getAttribute(String name) {
      return getAttribute(name, "");
   }

   /**
    * getBooleanAttribute returns an attribute from the Configuration node.
    *
    * @param name The attribute name
    * @param defValue The default value
    * @return The attribute value
    */
   public boolean getBooleanAttribute(String name, boolean defValue) {
      try {
         return "[true][yes][1][y][on]".indexOf("["+getAttribute(name, defValue ? "1" : "0").toLowerCase()+"]") != -1;
      }
      catch ( Exception e ) {
         return defValue;
      }
   }

   /**
    * getBooleanAttribute returns an attribute from the Configuration node.
    *
    * @param name The attribute name
    * @return The attribute value
    */
   public boolean getBooleanAttribute(String name) {
      return getBooleanAttribute(name, false);
   }

   /**
    * getIntAttribute returns an attribute from the Configuration node.
    *
    * @param name The attribute name
    * @param defValue The default value
    * @return The attribute value
    */
   public int getIntAttribute(String name, int defValue) {
      try {
         return Integer.parseInt(config.getAttribute(name));
      }
      catch ( Exception e ) {
         return defValue;
      }
   }

   /**
    * getIntAttribute returns an attribute from the Configuration node.
    *
    * @param name The attribute name
    * @return The attribute value
    */
   public int getIntAttribute(String name) {
      return getIntAttribute(name, 0);
   }

   /**
    * getShortAttribute returns an attribute from the Configuration node.
    *
    * @param name The attribute name
    * @param defValue The default value
    * @return The attribute value
    */
   public short getShortAttribute(String name, short defValue) {
      try {
         return Short.parseShort(config.getAttribute(name));
      }
      catch ( Exception e ) {
         return defValue;
      }
   }

   /**
    * getShortAttribute returns an attribute from the Configuration node.
    *
    * @param name The attribute name
    * @return The attribute value
    */
   public short getShortAttribute(String name) {
      return getShortAttribute(name, (short)0);
   }

   /**
    * getLongAttribute returns an attribute from the Configuration node.
    *
    * @param name The attribute name
    * @param defValue The default value
    * @return The attribute value
    */
   public long getLongAttribute(String name, long defValue) {
      try {
         return Long.parseLong(config.getAttribute(name));
      }
      catch ( Exception e ) {
         return defValue;
      }
   }

   /**
    * getLongAttribute returns an attribute from the Configuration node.
    *
    * @param name The attribute name
    * @return The attribute value
    */
   public long getLongAttribute(String name) {
      return getLongAttribute(name, (long)0);
   }

   /**
    * getFloatAttribute returns an attribute from the Configuration node.
    *
    * @param name The attribute name
    * @param defValue The default value
    * @return The attribute value
    */
   public float getFloatAttribute(String name, float defValue) {
      try {
         return Float.parseFloat(config.getAttribute(name));
      }
      catch ( Exception e ) {
         return defValue;
      }
   }

   /**
    * getFloatAttribute returns an attribute from the Configuration node.
    *
    * @param name The attribute name
    * @return The attribute value
    */
   public float getFloatAttribute(String name) {
      return getFloatAttribute(name, (float)0.0);
   }

   /**
    * getDoubleAttribute returns an attribute from the Configuration node.
    *
    * @param name The attribute name
    * @param defValue The default value
    * @return The attribute value
    */
   public double getDoubleAttribute(String name, double defValue) {
      try {
         return Double.parseDouble(config.getAttribute(name));
      }
      catch ( Exception e ) {
         return defValue;
      }
   }

   /**
    * getDoubleAttribute returns an attribute from the Configuration node.
    *
    * @param name The attribute name
    * @return The attribute value
    */
   public double getDoubleAttribute(String name) {
      return getDoubleAttribute(name, (double)0.0);
   }

   /**
    * getByteAttribute returns an attribute from the Configuration node.
    *
    * @param name The attribute name
    * @param defValue The default value
    * @return The attribute value
    */
   public byte getByteAttribute(String name, byte defValue) {
      try {
         return Byte.parseByte(config.getAttribute(name));
      }
      catch ( Exception e ) {
         return defValue;
      }
   }

   /**
    * getByteAttribute returns an attribute from the Configuration node.
    *
    * @param name The attribute name
    * @return The attribute value
    */
   public byte getByteAttribute(String name) {
      return getByteAttribute(name, (byte)0);
   }

   /**
    * getCharAttribute returns an attribute from the Configuration node.
    *
    * @param name The attribute name
    * @param defValue The default value
    * @return The attribute value
    */
   public char getCharAttribute(String name, char defValue) {
      try {
         return config.getAttribute(name).charAt(0);
      }
      catch ( Exception e ) {
         return defValue;
      }
   }

   /**
    * getCharAttribute returns an attribute from the Configuration node.
    *
    * @param name The attribute name
    * @return The attribute value
    */
   public char getCharAttribute(String name) {
      return getCharAttribute(name, '\0');
   }

   /**
    * listAttributes returns a list of the attribute names that exist for
    * this Configuration node.
    *
    * @return The Attribute names
    */
   public String[] listAttributes() {
      NamedNodeMap map = config.getAttributes();
      String[] result = new String[map.getLength()];
      int size = map.getLength();
      for ( int i = 0; i < size; i++ )
         result[i] = ((Attr)map.item(i)).getName();
      return result;
   }

   /**
    * hasValue returns whether or not the Configuration node has a textual
    * value associated with it.
    *
    * @return Whether or not the node has a text value
    */
   public boolean hasValue() {
      NodeList list = config.getChildNodes();
      return ( list.getLength() > 0 && list.item(0).getNodeType() == Node.TEXT_NODE );
   }

   /**
    * getValue returns the value from the Configuration node.
    *
    * @param defValue The default value
    * @return The value
    */
   public String getValue(String defValue) {
      NodeList list = config.getChildNodes();
      if ( list.getLength() == 1 && list.item(0).getNodeType() == Node.TEXT_NODE )
         return list.item(0).getNodeValue();
      else
         return defValue;
   }

   /**
    * getValue returns the value from the Configuration node.
    *
    * @return The value
    */
   public String getValue() {
      return getValue("");
   }

   /**
    * getBooleanValue returns the value from the Configuration node.
    *
    * @param defValue The default value
    * @return The value
    */
   public boolean getBooleanValue(String name, boolean defValue) {
      try {
         return "[true][yes][1][y][on]".indexOf("["+getValue().toLowerCase()+"]") != -1;
      }
      catch ( Exception e ) {
         return defValue;
      }
   }

   /**
    * getBooleanValue returns the value from the Configuration node.
    *
    * @return The value
    */
   public boolean getBooleanValue(String name) {
      return getBooleanValue(name, false);
   }

   /**
    * getIntValue returns the value from the Configuration node.
    *
    * @param defValue The default value
    * @return The value
    */
   public int getIntValue(String name, int defValue) {
      try {
         return Integer.parseInt(getValue());
      }
      catch ( Exception e ) {
         return defValue;
      }
   }

   /**
    * getIntValue returns the value from the Configuration node.
    *
    * @return The value
    */
   public int getIntValue(String name) {
      return getIntValue(name, 0);
   }

   /**
    * getShortValue returns the value from the Configuration node.
    *
    * @param defValue The default value
    * @return The value
    */
   public short getShortValue(String name, short defValue) {
      try {
         return Short.parseShort(getValue());
      }
      catch ( Exception e ) {
         return defValue;
      }
   }

   /**
    * getShortValue returns the value from the Configuration node.
    *
    * @return The value
    */
   public short getShortValue(String name) {
      return getShortValue(name, (short)0);
   }

   /**
    * getLongValue returns the value from the Configuration node.
    *
    * @param defValue The default value
    * @return The value
    */
   public long getLongValue(String name, long defValue) {
      try {
         return Long.parseLong(getValue());
      }
      catch ( Exception e ) {
         return defValue;
      }
   }

   /**
    * getLongValue returns the value from the Configuration node.
    *
    * @return The value
    */
   public long getLongValue(String name) {
      return Long.parseLong(getValue());
   }

   /**
    * getFloatValue returns the value from the Configuration node.
    *
    * @param defValue The default value
    * @return The value
    */
   public float getFloatValue(String name, float defValue) {
      try {
         return Float.parseFloat(getValue());
      }
      catch ( Exception e ) {
         return defValue;
      }
   }

   /**
    * getFloatValue returns the value from the Configuration node.
    *
    * @return The value
    */
   public float getFloatValue(String name) {
      return getFloatValue(name, (float)0.0);
   }

   /**
    * getDoubleValue returns the value from the Configuration node.
    *
    * @param defValue The default value
    * @return The value
    */
   public double getDoubleValue(String name, double defValue) {
      try {
         return Double.parseDouble(getValue());
      }
      catch ( Exception e ) {
         return defValue;
      }
   }

   /**
    * getDoubleValue returns the value from the Configuration node.
    *
    * @return The value
    */
   public double getDoubleValue(String name) {
      return getDoubleValue(name, (double)0.0);
   }

   /**
    * getByteValue returns the value from the Configuration node.
    *
    * @param defValue The default value
    * @return The value
    */
   public byte getByteValue(String name, byte defValue) {
      try {
         return Byte.parseByte(getValue());
      }
      catch ( Exception e ) {
         return defValue;
      }
   }

   /**
    * getByteValue returns the value from the Configuration node.
    *
    * @return The value
    */
   public byte getByteValue(String name) {
      return getByteValue(name, (byte)0);
   }

   /**
    * getCharValue returns the value from the Configuration node.
    *
    * @param defValue The default value
    * @return The value
    */
   public char getCharValue(String name, char defValue) {
      try {
         return getValue().charAt(0);
      }
      catch ( Exception e ) {
         return defValue;
      }
   }

   /**
    * getCharValue returns the value from the Configuration node.
    *
    * @return The value
    */
   public char getCharValue(String name) {
      return getCharValue(name, '\0');
   }

   /**
    * hasChildren returns whether or not the Configuration node has child
    * Configuration elements.
    *
    * @return Whether or not children exist
    */
   public boolean hasChildren() {
      NodeList list = config.getChildNodes();
      int size = list.getLength();
      for ( int i = 0; i < size; i++ )
         if ( list.item(i).getNodeType() == Node.ELEMENT_NODE )
            return true;
      return false;
   }

   /**
    * getChildren returns the child Configuration nodes for this Configuration
    * node.
    *
    * @return The children
    */
   public Configuration[] getChildren() {
      NodeList list = config.getChildNodes();
      if ( list == null )
         return EmptySet;
      List tmp = new ArrayList();
      int size = list.getLength();
      for ( int i = 0; i < size; i++ )
         if ( list.item(i).getNodeType() == Node.ELEMENT_NODE )
            tmp.add(new Configuration((Element)list.item(i), readOnly));
      return (Configuration[])tmp.toArray(EmptySet);
   }

   /**
    * processChildren runs through the children of this Configuration and
    * performs prcessing callbacks to the ConfigurationCallback object.
    *
    * @param callback The callback object
    */
   public void processChildren(ConfigurationCallback callback) {
      NodeList list = config.getChildNodes();
      if ( list == null )
         return;
      int size = list.getLength();
      try {
         for ( int i = 0; i < size; i++ )
            if ( list.item(i).getNodeType() == Node.ELEMENT_NODE )
               callback.process(new Configuration((Element)list.item(i), readOnly));
      }
      catch (Exception e) {
         org.apache.xindice.Debug.printStackTrace(e);
      }
   }

   /**
    * processChildren runs through the children of this Configuration and
    * performs prcessing callbacks to the ConfigurationCallback object.
    *
    * @param callback The callback object
    * @param name The Configuration name to process
    */
   public void processChildren(String name, ConfigurationCallback callback) {
      NodeList list = config.getChildNodes();
      if ( list == null )
         return;
      
      Node[] nl = new Node[list.getLength()];
      for ( int i = 0; i < nl.length; i++ )
         nl[i] = list.item(i);

      int size = nl.length;
      
      try {
         for ( int i = 0; i < size; i++ )
            if ( nl[i].getNodeType() == Node.ELEMENT_NODE && nl[i].getNodeName().equals(name))
               callback.process(new Configuration((Element)nl[i], readOnly));
      }
      catch (Exception e) {
         org.apache.xindice.Debug.printStackTrace(e);
      }
   }

   /**
    * getChild returns the first child Configuration node that matches the
    * specified name.
    *
    * @param name The Configuration node name
    * @param create Create Configuration if it doesn't exist
    * @return The child Configuration
    * @throws ReadOnlyException if the Configuration is read-only
    */
   public Configuration getChild(String name, boolean create) throws ReadOnlyException {
      NodeList list = config.getChildNodes();
      Configuration[] result = new Configuration[list.getLength()];
      int size = list.getLength();
         
      for ( int i = 0; i < size; i++ )
         if ( list.item(i).getNodeType() == Node.ELEMENT_NODE && list.item(i).getNodeName().equals(name) )
            return new Configuration((Element)list.item(i), readOnly);
      if ( create )
         return add(name);
      else
         return null;
   }

   /**
    * getChild returns the first child Configuration node that matches the
    * specified name.
    *
    * @param name The Configuration node name
    * @return The child Configuration
    */
   public Configuration getChild(String name) {
      try {
         return getChild(name, false);
      }
      catch( Exception e ) {
         return null;
      }
   }

   /**
    * isReadOnly returns whether or not this Configuration node is a
    * read-only node.
    *
    * @return The read-only status
    */
   public boolean isReadOnly() {
      return readOnly;
   }

   /**
    * add adds a new Configuration node to this Configuration node.
    *
    * @param name The name of the Configuration node
    * @return The new Configuration node
    * @throws ReadOnlyException if the Configuration is read-only
    */
   public Configuration add(String name) throws ReadOnlyException {
      if ( readOnly )
         throw new ReadOnlyException();
      Element elem = config.getOwnerDocument().createElement(name);
      config.appendChild(elem);
      setDirty();
      return new Configuration(elem, readOnly);
   }

   /**
    * add adds an existing Configuration node to this Configuration node.
    * Note!  This method does NOT perform a deep copy on the DOM Node that
    * newConfig manages.
    *
    * @param config The Configuration node to add
    * @throws ReadOnlyException if the Configuration is read-only
    */
   public void add(Configuration newConfig) throws ReadOnlyException {
      if ( readOnly )
         throw new ReadOnlyException();
      Node imported = config.getOwnerDocument().importNode(newConfig.config, true);
      config.appendChild(imported);
      newConfig.config = (Element)imported;
      setDirty();
   }

   /**
    * delete deletes this Configuration node from its parent Configuration.
    *
    * @throws ReadOnlyException if the Configuration is read-only
    */
   public void delete() throws ReadOnlyException {
      if ( readOnly )
         throw new ReadOnlyException();
      if ( config.getParentNode() != null )
         config.getParentNode().removeChild(config);
      setDirty();
   }

   /**
    * removeAttribute removes the named attribute from this Configuration.
    *
    * @param name The attribute name
    * @throws ReadOnlyException if the Configuration is read-only
    */
   public void removeAttribute(String name) throws ReadOnlyException {
      if ( readOnly )
         throw new ReadOnlyException();
      config.removeAttribute(name);
      setDirty();
   }

   /**
    * setAttribute sets the attribute value for this Configuration.
    *
    * @param name The attribute name
    * @param value The attribute value
    * @throws ReadOnlyException if the Configuration is read-only
    */
   public void setAttribute(String name, String value) throws ReadOnlyException {
      if ( readOnly )
         throw new ReadOnlyException();
      config.setAttribute(name, value);
      setDirty();
   }

   /**
    * setAttribute sets the attribute value for this Configuration.
    *
    * @param name The attribute name
    * @param value The attribute value
    * @throws ReadOnlyException if the Configuration is read-only
    */
   public void setAttribute(String name, boolean value) throws ReadOnlyException {
      if ( readOnly )
         throw new ReadOnlyException();
      config.setAttribute(name, String.valueOf(value));
      setDirty();
   }

   /**
    * setAttribute sets the attribute value for this Configuration.
    *
    * @param name The attribute name
    * @param value The attribute value
    * @throws ReadOnlyException if the Configuration is read-only
    */
   public void setAttribute(String name, int value) throws ReadOnlyException {
      if ( readOnly )
         throw new ReadOnlyException();
      config.setAttribute(name, String.valueOf(value));
      setDirty();
   }

   /**
    * setAttribute sets the attribute value for this Configuration.
    *
    * @param name The attribute name
    * @param value The attribute value
    * @throws ReadOnlyException if the Configuration is read-only
    */
   public void setAttribute(String name, short value) throws ReadOnlyException {
      if ( readOnly )
         throw new ReadOnlyException();
      config.setAttribute(name, String.valueOf(value));
      setDirty();
   }

   /**
    * setAttribute sets the attribute value for this Configuration.
    *
    * @param name The attribute name
    * @param value The attribute value
    * @throws ReadOnlyException if the Configuration is read-only
    */
   public void setAttribute(String name, long value) throws ReadOnlyException {
      if ( readOnly )
         throw new ReadOnlyException();
      config.setAttribute(name, String.valueOf(value));
      setDirty();
   }

   /**
    * setAttribute sets the attribute value for this Configuration.
    *
    * @param name The attribute name
    * @param value The attribute value
    * @throws ReadOnlyException if the Configuration is read-only
    */
   public void setAttribute(String name, float value) throws ReadOnlyException {
      if ( readOnly )
         throw new ReadOnlyException();
      config.setAttribute(name, String.valueOf(value));
      setDirty();
   }

   /**
    * setAttribute sets the attribute value for this Configuration.
    *
    * @param name The attribute name
    * @param value The attribute value
    * @throws ReadOnlyException if the Configuration is read-only
    */
   public void setAttribute(String name, double value) throws ReadOnlyException {
      if ( readOnly )
         throw new ReadOnlyException();
      config.setAttribute(name, String.valueOf(value));
      setDirty();
   }

   /**
    * setAttribute sets the attribute value for this Configuration.
    *
    * @param name The attribute name
    * @param value The attribute value
    * @throws ReadOnlyException if the Configuration is read-only
    */
   public void setAttribute(String name, byte value) throws ReadOnlyException {
      if ( readOnly )
         throw new ReadOnlyException();
      config.setAttribute(name, String.valueOf(value));
      setDirty();
   }

   /**
    * setAttribute sets the attribute value for this Configuration.
    *
    * @param name The attribute name
    * @param value The attribute value
    * @throws ReadOnlyException if the Configuration is read-only
    */
   public void setAttribute(String name, char value) throws ReadOnlyException {
      if ( readOnly )
         throw new ReadOnlyException();
      config.setAttribute(name, String.valueOf(value));
      setDirty();
   }

   /**
    * setValue sets the value of the Configuration node.
    *
    * @param value The New Value
    * @throws ReadOnlyException if the Configuration is read-only
    */
   public void setValue(String value) throws ReadOnlyException {
      if ( readOnly )
         throw new ReadOnlyException();
      NodeList list = config.getChildNodes();
      if ( list.getLength() == 1 && list.item(0).getNodeType() == Node.TEXT_NODE )
         list.item(0).setNodeValue(value);
      else {
         Text text = config.getOwnerDocument().createTextNode(value);
         config.appendChild(text);
      }
      setDirty();
   }

   /**
    * setValue sets the value of the Configuration node.
    *
    * @param value The New Value
    * @throws ReadOnlyException if the Configuration is read-only
    */
   public void setValue(boolean value) throws ReadOnlyException {
      setValue( value ? "true"
                      : "false" );
   }

   /**
    * setValue sets the value of the Configuration node.
    *
    * @param value The New Value
    * @throws ReadOnlyException if the Configuration is read-only
    */
   public void setValue(int value) throws ReadOnlyException {
      setValue(String.valueOf(value));
   }

   /**
    * setValue sets the value of the Configuration node.
    *
    * @param value The New Value
    * @throws ReadOnlyException if the Configuration is read-only
    */
   public void setValue(short value) throws ReadOnlyException {
      setValue(String.valueOf(value));
   }

   /**
    * setValue sets the value of the Configuration node.
    *
    * @param value The New Value
    * @throws ReadOnlyException if the Configuration is read-only
    */
   public void setValue(long value) throws ReadOnlyException {
      setValue(String.valueOf(value));
   }

   /**
    * setValue sets the value of the Configuration node.
    *
    * @param value The New Value
    * @throws ReadOnlyException if the Configuration is read-only
    */
   public void setValue(float value) throws ReadOnlyException {
      setValue(String.valueOf(value));
   }

   /**
    * setValue sets the value of the Configuration node.
    *
    * @param value The New Value
    * @throws ReadOnlyException if the Configuration is read-only
    */
   public void setValue(double value) throws ReadOnlyException {
      setValue(String.valueOf(value));
   }

   /**
    * setValue sets the value of the Configuration node.
    *
    * @param value The New Value
    * @throws ReadOnlyException if the Configuration is read-only
    */
   public void setValue(byte value) throws ReadOnlyException {
      setValue(String.valueOf(value));
   }

   /**
    * setValue sets the value of the Configuration node.
    *
    * @param value The New Value
    * @throws ReadOnlyException if the Configuration is read-only
    */
   public void setValue(char value) throws ReadOnlyException {
      setValue(String.valueOf(value));
   }
}

